# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jwskate',
 'jwskate.jwa',
 'jwskate.jwa.encryption',
 'jwskate.jwa.key_mgmt',
 'jwskate.jwa.signature',
 'jwskate.jwe',
 'jwskate.jwk',
 'jwskate.jws',
 'jwskate.jwt',
 'tests',
 'tests.test_jwk']

package_data = \
{'': ['*']}

install_requires = \
['backports.cached-property>=1.0.1,<2.0.0',
 'binapy>=0.5',
 'cryptography>=3.4',
 'typing-extensions>=4.2.0,<5.0.0']

setup_kwargs = {
    'name': 'jwskate',
    'version': '0.2.1',
    'description': 'A Pythonic implementation of Json Web Signature, Keys, Algorithms, Tokens and Encryption (RFC7514 to 7519), on top of the `cryptography` module.',
    'long_description': '# JwSkate\n\n[![PyPi](https://img.shields.io/pypi/v/jwskate.svg)](https://pypi.python.org/pypi/jwskate)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nA Pythonic implementation of Json Web Signature, Keys, Algorithms, Tokens and Encryption (RFC7514 to 7519), and their\nextensions ECDH Signatures (RFC8037), and JWK Thumbprints (RFC7638).\n\n- Free software: MIT\n- Documentation: <https://guillp.github.io/jwskate/>\n\nA quick usage example, generating an RSA private key, signing some data, then validating that signature:\n\n```python\nfrom jwskate import Jwk\n\n# generate a RSA Jwk and sign a plaintext with it\nrsa_private_jwk = Jwk.generate_for_kty("RSA", key_size=2048, kid="my_key")\n\ndata = b"Signing is easy!"\nalg = "RS256"\nsignature = rsa_private_jwk.sign(data, alg)\n\n# extract the public key, and verify the signature with it\nrsa_public_jwk = rsa_private_jwk.public_jwk()\nassert rsa_public_jwk.verify(data, signature, alg)\n\n# let\'s see what a Jwk looks like:\nassert isinstance(rsa_private_jwk, dict)  # Jwk are dict\n\nprint(rsa_private_jwk)\n```\n\nThe result of this print JWK will look like this:\n\n```\n{ \'kty\': \'RSA\',\n  \'n\': \'...\',\n  \'e\': \'AQAB\',\n  \'d\': \'...\',\n  \'p\': \'...\',\n  \'q\': \'...\',\n  \'dp\': \'...\',\n  \'dq\': \'...\',\n  \'qi\': \'...\',\n  \'kid\': \'my_key\'\n}\n```\n\nNow let\'s sign a JWT containing arbitrary claims:\n\n```python\nfrom jwskate import Jwk, Jwt\n\nprivate_jwk = Jwk.generate_for_kty("EC", kid="my_key")\nclaims = {"sub": "some_sub", "claim1": "value1"}\nsign_alg = "ES256"\n\njwt = Jwt.sign(claims, private_jwk, sign_alg)\n# that\'s it! we have a signed JWT\nassert jwt.claims == claims  # claims can be accessed as a dict\nassert jwt.sub == "some_sub"  # or individual claims can be accessed as attributes\nassert jwt["claim1"] == "value1"  # or as dict items\nassert jwt.alg == sign_alg  # alg and kid headers are also accessible as attributes\nassert jwt.kid == private_jwk.kid\nassert jwt.verify_signature(private_jwk.public_jwk(), sign_alg)\n\nprint(jwt)\n```\n\nThis will output the full JWT compact representation. You can inspect it for example at <https://jwt.io>\n\n```\neyJhbGciOiJFUzI1NiIsImtpZCI6Im15a2V5In0.eyJzdWIiOiJzb21lX3N1YiIsImNsYWltMSI6InZhbHVlMSJ9.C1KcDyDT8qXwUqcWzPKkQD7f6xai-gCgaRFMdKPe80Vk7XeYNa8ovuLwvdXgGW4ZZ_lL73QIyncY7tHGXUthag\n```\n\nOr let\'s sign a JWT with the standardised lifetime, subject, audience and ID claims:\n\n```python\nfrom jwskate import Jwk, JwtSigner\n\nprivate_jwk = Jwk.generate_for_kty("EC")\nsigner = JwtSigner(issuer="https://myissuer.com", jwk=private_jwk, alg="ES256")\njwt = signer.sign(\n    subject="some_sub",\n    audience="some_aud",\n    extra_claims={"custom_claim1": "value1", "custom_claim2": "value2"},\n)\n\nprint(jwt.claims)\n```\n\nThe generated JWT claims will include the standardised claims:\n\n```\n{\'custom_claim1\': \'value1\',\n \'custom_claim2\': \'value2\',\n \'iss\': \'https://myissuer.com\',\n \'aud\': \'some_aud\',\n \'sub\': \'some_sub\',\n \'iat\': 1648823184,\n \'exp\': 1648823244,\n \'jti\': \'3b400e27-c111-4013-84e0-714acd76bf3a\'\n}\n```\n\n## Features\n\n- Simple, Clean, Pythonic interface\n- Convenience wrappers around `cryptography` for all algorithms described in JWA\n- Json Web Keys (JWK) loading and generation\n- Arbitrary data signature and verification using Json Web Keys\n- Json Web Signatures (JWS) signing and verification\n- Json Web Encryption (JWE) encryption and decryption\n- Json Web Tokens (JWT) signing, verification and validation\n- 100% type annotated\n- nearly 100% code coverage\n- Relies on [cryptography](https://cryptography.io) for all cryptographic operations\n- Relies on [BinaPy](https://guillp.github.io/binapy/) for binary data manipulations\n\n### Supported Signature algorithms\n\n`jwskate` supports the following signature algorithms:\n\n| Signature Alg | Description                                    | Key Type | Reference                          | Note                  |\n|---------------|------------------------------------------------|----------|------------------------------------|-----------------------|\n| HS256         | HMAC using SHA-256                             | oct      | RFC7518, Section 3.2               |                       |\n| HS384         | HMAC using SHA-384                             | oct      | RFC7518, Section 3.2               |                       |\n| HS512         | HMAC using SHA-512                             | oct      | RFC7518, Section 3.2               |                       |\n| RS256         | RSASSA-PKCS1-v1_5 using SHA-256                | RSA      | RFC7518, Section 3.3               |                       |\n| RS384         | RSASSA-PKCS1-v1_5 using SHA-384                | RSA      | RFC7518, Section 3.3               |                       |\n| RS512         | RSASSA-PKCS1-v1_5 using SHA-512                | RSA      | RFC7518, Section 3.3               |                       |\n| ES256         | ECDSA using P-256 and SHA-256                  | EC       | RFC7518, Section 3.4               |                       |\n| ES384         | ECDSA using P-384 and SHA-384                  | EC       | RFC7518, Section 3.4               |                       |\n| ES512         | ECDSA using P-521 and SHA-512                  | EC       | RFC7518, Section 3.4               |                       |\n| PS256         | RSASSA-PSS using SHA-256 and MGF1 with SHA-256 | RSA      | RFC7518, Section 3.5               |                       |\n| PS384         | RSASSA-PSS using SHA-384 and MGF1 with SHA-384 | RSA      | RFC7518, Section 3.5               |                       |\n| PS512         | RSASSA-PSS using SHA-512 and MGF1 with SHA-512 | RSA      | RFC7518, Section 3.5               |                       |\n| EdDSA         | EdDSA signature algorithms                     | OKP      | RFC8037, Section 3.1               |                       |\n| ES256K        | ECDSA using secp256k1 curve and SHA-256        | EC       | RFC8812, Section 3.2               |                       |\n| HS1           | HMAC using SHA-1                               | oct      | https://www.w3.org/TR/WebCryptoAPI | Validation Only       |\n| RS1           | RSASSA-PKCS1-v1_5 with SHA-1                   | oct      | https://www.w3.org/TR/WebCryptoAPI | Validation Only       |\n| none          | No digital signature or MAC performed          |          | RFC7518, Section 3.6               | Not usable by mistake |\n\n### Supported Key Management algorithms\n\n`jwskate` supports the following key management algorithms:\n\n| Signature Alg      | Description                                     | Key Type | Reference                          | Note        |\n|--------------------|-------------------------------------------------|----------|------------------------------------|-------------|\n| RSA1_5             | RSAES-PKCS1-v1_5                                | RSA      | RFC7518, Section 4.2               | Unwrap Only |\n| RSA-OAEP           | RSAES OAEP using default parameters             | RSA      | RFC7518, Section 4.3               |             |\n| RSA-OAEP-256       | RSAES OAEP using SHA-256 and MGF1 with SHA-256  | RSA      | RFC7518, Section 4.3               |             |\n| RSA-OAEP-384       | RSA-OAEP using SHA-384 and MGF1 with SHA-384    | RSA      | https://www.w3.org/TR/WebCryptoAPI |             |\n| RSA-OAEP-512       | RSA-OAEP using SHA-512 and MGF1 with SHA-512    | RSA      | https://www.w3.org/TR/WebCryptoAPI |             |\n| A128KW             | AES Key Wrap using 128-bit key                  | oct      | RFC7518, Section 4.4               |             |\n| A192KW             | AES Key Wrap using 192-bit key                  | oct      | RFC7518, Section 4.4               |             |\n| A256KW             | AES Key Wrap using 256-bit key                  | oct      | RFC7518, Section 4.4               |             |\n| dir                | Direct use of a shared symmetric key            | oct      | RFC7518, Section 4.5               |             |\n| ECDH-ES            | ECDH-ES using Concat KDF                        | EC       | RFC7518, Section 4.6               |             |\n| ECDH-ES+A128KW     | ECDH-ES using Concat KDF and "A128KW" wrapping  | EC       | RFC7518, Section 4.6               |             |\n| ECDH-ES+A192KW     | ECDH-ES using Concat KDF and "A192KW" wrapping  | EC       | RFC7518, Section 4.6               |             |\n| ECDH-ES+A256KW     | ECDH-ES using Concat KDF and "A256KW" wrapping  | EC       | RFC7518, Section 4.6               |             |\n| A128GCMKW          | Key wrapping with AES GCM using 128-bit key     | oct      | RFC7518, Section 4.7               |             |\n| A192GCMKW          | Key wrapping with AES GCM using 192-bit key     | oct      | RFC7518, Section 4.7               |             |\n| A256GCMKW          | Key wrapping with AES GCM using 256-bit key     | oct      | RFC7518, Section 4.7               |             |\n| PBES2-HS256+A128KW | PBES2 with HMAC SHA-256 and "A128KW" wrapping   | password | RFC7518, Section 4.8               |             |\n| PBES2-HS384+A192KW | PBES2 with HMAC SHA-384 and "A192KW" wrapping   | password | RFC7518, Section 4.8               |             |\n| PBES2-HS512+A256KW | PBES2 with HMAC SHA-512 and "A256KW" wrapping   | password | RFC7518, Section 4.8               |             |\n\n### Supported Encryption algorithms\n\n`jwskate` supports the following encryption algorithms:\n\n| Signature Alg  | Description                                                 | Reference               |\n|----------------|-------------------------------------------------------------|-------------------------|\n| A128CBC-HS256  | AES_128_CBC_HMAC_SHA_256 authenticated encryption algorithm | RFC7518, Section 5.2.3  |\n| A192CBC-HS384  | AES_192_CBC_HMAC_SHA_384 authenticated encryption algorithm | RFC7518, Section 5.2.4  |\n| A256CBC-HS512  | AES_256_CBC_HMAC_SHA_512 authenticated encryption algorithm | RFC7518, Section 5.2.5  |\n| A128GCM        | AES GCM using 128-bit key                                   | RFC7518, Section 5.3    |\n| A192GCM        | AES GCM using 192-bit key                                   | RFC7518, Section 5.3    |\n| A256GCM        | AES GCM using 256-bit key                                   | RFC7518, Section 5.3    |\n\n### Supported Elliptic Curves\n\n`jwskate` supports the following Elliptic Curves:\n\n| Curve     | Description                           | Key Type | Usage                 | Reference                |\n|-----------|---------------------------------------|----------|-----------------------|--------------------------|\n| P-256     | P-256 Curve                           | EC       | signature, encryption | RFC7518, Section 6.2.1.1 |\n| P-384     | P-384 Curve                           | EC       | signature, encryption | RFC7518, Section 6.2.1.1 |\n| P-521     | P-521 Curve                           | EC       | signature, encryption | RFC7518, Section 6.2.1.1 |\n| secp256k1 | SECG secp256k1 curve                  | EC       | signature, encryption | RFC8812, Section 3.1     |\n| Ed25519   | Ed25519 signature algorithm key pairs | OKP      | signature             | RFC8037, Section 3.1     |\n| Ed448     | Ed448 signature algorithm key pairs   | OKP      | signature             | RFC8037, Section 3.1     |\n| X25519    | X25519 function key pairs             | OKP      | encryption            | RFC8037, Section 3.2     |\n| X448      | X448 function key pairs               | OKP      | encryption            | RFC8037, Section 3.2     |\n\n## Why a new lib ?\n\nThere are already multiple implementations of JOSE and Json Web Crypto related specifications in Python. However, I have\nbeen dissatisfied by all of them so far, so I decided to come up with my own module.\n\n- [PyJWT](https://pyjwt.readthedocs.io): lacks support for JWK, JWE, JWS, requires keys in PEM format.\n- [JWCrypto](https://jwcrypto.readthedocs.io/): very inconsistent and complex API.\n- [Python-JOSE](https://python-jose.readthedocs.io/): lacks easy support for JWT validation\n  (checking the standard claims like iss, exp, etc.), lacks easy access to claims\n\nNot to say that those are _bad_ libs (I actually use `jwcrypto` myself for `jwskate` unit tests), but they either don\'t\nsupport some important features, or they just don\'t feel easy-enough, Pythonic-enough to use.\n\n## Design\n\n### JWK are dicts\n\nJWK are specified as JSON objects, which are parsed as `dict` in Python. The `Jwk` class in `jwskate` is actually a\n`dict` subclass, so you can use it exactly like you would use a dict: you can access its members, dump it back as JSON, etc.\nThe same is true for Json Web tokens in JSON format.\n\n### JWA Wrappers\n\nWhile you can directly use `cryptography` to do the cryptographic operations that are described in [JWA](https://www.rfc-editor.org/info/rfc7518),\nits usage is not straightforward and gives you plenty of options to carefully select, leaving room for errors.\nTo work around this, `jwskate` comes with a set of wrappers that implement the exact JWA specification, with minimum\nrisk of mistakes.\n\n### Safe Signature Verification\n\nFor every signature verification method in `jwskate`, you have to provide the expected signature(s) algorithm(s).\nThat is to avoid a security flaw where your application accepts tokens with a weaker encryption scheme than what\nyour security policy mandates; or even worse, where it accepts unsigned tokens, or tokens that are symmetrically signed\nwith an improperly used public key, leaving your application exposed to exploitation by attackers.\n\nEach signature verification accepts 2 args `alg` and `algs`. If you always expect to verify tokens signed with a single\nsignature algorithm, pass that algorithm ID to alg. If you accept multiple algs (for example, any asymmetric alg that\nyou consider strong enough), you can instead pass an iterable of allowed algorithms with `algs`. The signature will be\nvalidated as long as it is signed with one of the provided algs.\n\nFor verification methods that accept a `Jwk` key, you don\'t have to provide an `alg` or `algs` if that Jwk has the\nappropriate `alg` member that define which algorithm is supposed to be used with that key.\n\n## Credits\n\nAll cryptographic operations are handled by [cryptography](https://cryptography.io).\n',
    'author': 'Guillaume Pujol',
    'author_email': 'guill.p.linux@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/guillp/jwskate',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""This module implements the `Jwk` base class, which provides most of the common features of all JWK types."""

from __future__ import annotations

import hashlib
import json
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    Union,
)

from binapy import BinaPy
from cryptography.hazmat.primitives import serialization

from jwskate.jwa import (
    BaseAESEncryptionAlg,
    BaseAesGcmKeyWrap,
    BaseAesKeyWrap,
    BaseAsymmetricAlg,
    BaseEcdhEs_AesKw,
    BaseKeyManagementAlg,
    BaseRsaKeyWrap,
    BaseSignatureAlg,
    BaseSymmetricAlg,
    DirectKeyUse,
    EcdhEs,
)

from ..token import BaseJsonDict
from .alg import UnsupportedAlg, select_alg, select_algs

if TYPE_CHECKING:
    from .jwks import JwkSet  # pragma: no cover


class UnsupportedKeyType(ValueError):
    """Raised when an unsupported Key Type is requested."""


class InvalidJwk(ValueError):
    """Raised when an invalid JWK is encountered."""


@dataclass
class JwkParameter:  # noqa: D101
    description: str
    is_private: bool
    is_required: bool
    kind: str


class Jwk(BaseJsonDict):
    """Represents a Json Web Key (JWK), as specified in RFC7517.

    A JWK is a JSON object that represents a cryptographic key.  The
    members of the object represent properties of the key, including its
    value. Just like a parsed JSON object, a :class:`Jwk` is a dict, so
    you can do with a Jwk anything you can do with a `dict`. In
    addition, all keys parameters are exposed as attributes. There are
    subclasses of `Jwk` for each specific Key Type, but you shouldn't
    have to use the subclasses directly since they all present a common
    interface.
    """

    subclasses: Dict[str, Type[Jwk]] = {}
    """A dict of 'kty' values to subclasses implementing each specific Key Type"""

    cryptography_key_types: Dict[Any, Type[Jwk]] = {}
    """A dict of cryptography key classes to its specific 'kty' value"""

    PARAMS: Mapping[str, JwkParameter]
    """A dict of parameters. Key is parameter name, value is a tuple (description, is_private, is_required, kind)"""

    KTY: ClassVar[str]
    """The Key Type associated with this JWK."""

    CRYPTOGRAPHY_KEY_CLASSES: ClassVar[Iterable[Any]]

    SIGNATURE_ALGORITHMS: Mapping[str, Type[BaseSignatureAlg]] = {}
    KEY_MANAGEMENT_ALGORITHMS: Mapping[str, Type[BaseKeyManagementAlg]] = {}
    ENCRYPTION_ALGORITHMS: Mapping[str, Type[BaseAESEncryptionAlg]] = {}

    def __init_subclass__(cls) -> None:
        """Automatically add subclasses to the registry.

        This allows __new__ to pick the appropriate subclass when
        creating a Jwk
        """
        Jwk.subclasses[cls.KTY] = cls
        for klass in cls.CRYPTOGRAPHY_KEY_CLASSES:
            Jwk.cryptography_key_types[klass] = cls

    def __new__(cls, key: Union[Jwk, Dict[str, Any], Any], **kwargs: Any):  # type: ignore
        """Overridden `__new__` to make the Jwk constructor smarter.

        The Jwk constructor will accept:

            - a `dict` with the parsed Jwk content
            - another Jwk, which will be used as-is instead of creating a copy
            - an instance from a `cryptography` public or private key class

        Args:
            key: a dict containing JWK parameters, or another Jwk instance, or a `cryptography` key
            **kwargs: additional members to include in the Jwk
        """
        if cls == Jwk:
            if isinstance(key, Jwk):
                return cls.from_cryptography_key(key.cryptography_key, **kwargs)
            if isinstance(key, dict):
                kty: Optional[str] = key.get("kty")
                if kty is None:
                    raise ValueError("A Json Web Key must have a Key Type (kty)")

                subclass = Jwk.subclasses.get(kty)
                if subclass is None:
                    raise ValueError("Unsupported Key Type", kty)
                return super().__new__(subclass)
            elif isinstance(key, str):
                return cls.from_json(key)
            else:
                return cls.from_cryptography_key(key, **kwargs)
        return super().__new__(cls, key, **kwargs)

    def __init__(
        self, params: Union[Dict[str, Any], Any], include_kid_thumbprint: bool = False
    ):
        """Initialize a Jwk.

        This accepts a `dict` with the parsed Jwk contents, and an optional kid if it isn't already part of the dict. If no `kid` is supplied and `include_kid_thumbprint`, a default kid is generated based on the key thumbprint (defined in RFC7638).

        Args:
            params: a dict with the parsed Jwk parameters
            include_kid_thumbprint: if `True` (default), and there is no kid in the provided params, generate a kid based on the key thumbprint
        """
        if isinstance(
            params, dict
        ):  # this is to avoid double init due to the __new__ above
            super().__init__(
                {key: val for key, val in params.items() if val is not None}
            )
            self._validate()
            if self.get("kid") is None and include_kid_thumbprint:
                self["kid"] = self.thumbprint()

        try:
            self.cryptography_key = self._to_cryptography_key()
        except AttributeError as exc:
            raise InvalidJwk() from exc

    @property
    def is_private(self) -> bool:
        """Return `True` if the key is private, `False` otherwise.

        Returns:
            `True` if the key is private, `False` otherwise
        """
        return True

    def __getattr__(self, item: str) -> Any:
        """Allows access to key parameters as attributes, like `jwk.kid`, `jwk.kty`, instead of `jwk['kid']`, `jwk['kty']`, etc.

        Args:
            item: the member to access

        Return:
            the member value

        Raises:
            AttributeError: if the member is not found
        """
        value = self.get(item)
        if value is None:
            raise AttributeError(item)
        return value

    def __setitem__(self, key: str, value: Any) -> None:
        """Override base method to avoid modifying cryptographic key attributes.

        Args:
            key: name of the attribute to set
            value: value to set

        Raises:
            RuntimeError: when trying to modify cryptographic attributes

        """
        if key in self.PARAMS:
            raise RuntimeError("JWK key attributes cannot be modified.")
        super().__setitem__(key, value)

    def thumbprint(self, hashalg: str = "SHA256") -> str:
        """Return the key thumbprint as specified by RFC 7638.

        Args:
          hashalg: A hash function (defaults to SHA256)

        Returns:
            the calculated thumbprint
        """
        digest = hashlib.new(hashalg)

        t = {"kty": self.get("kty")}
        for name, param in self.PARAMS.items():
            if param.is_required and not param.is_private:
                t[name] = self.get(name)

        intermediary = json.dumps(t, separators=(",", ":"), sort_keys=True)
        digest.update(intermediary.encode("utf8"))
        return BinaPy(digest.digest()).to("b64u").ascii()

    @property
    def kty(self) -> str:
        """Return the Key Type.

        Returns:
            the key type
        """
        return self.KTY

    @property
    def alg(self) -> Optional[str]:
        """Return the configured key alg, if any.

        Returns:
            the key alg
        """
        alg = self.get("alg")
        if alg is not None and not isinstance(alg, str):
            raise TypeError(f"Invalid alg type {type(str)}", alg)
        return alg

    def _validate(self) -> None:
        """Internal method used to validate a Jwk. It checks that all required parameters are present and well-formed. If the key is private, it sets the `is_private` flag to `True`.

        Raises:
            TypeError: if the key type doesn't match the subclass
            InvalidJwk: if the JWK misses required members or has invalid members
        """
        if self.get("kty") != self.KTY:
            raise TypeError(
                f"This key 'kty' {self.get('kty')} doesn't match this Jwk subclass intended 'kty' {self.KTY}!"
            )

        jwk_is_private = False
        for name, param in self.PARAMS.items():

            value = self.get(name)

            if param.is_private and value is not None:
                jwk_is_private = True

            if not param.is_private and param.is_required and value is None:
                raise InvalidJwk(
                    f"Missing required public param {param.description} ({name})"
                )

            if value is None:
                pass
            elif param.kind == "b64u":
                if not isinstance(value, str):
                    raise InvalidJwk(
                        f"Parameter {param.description} ({name}) must be a string with a Base64URL-encoded value"
                    )
                if not BinaPy(value).check("b64u"):
                    raise InvalidJwk(
                        f"Parameter {param.description} ({name}) must be a Base64URL-encoded value"
                    )
            elif param.kind == "unsupported":
                if value is not None:
                    raise InvalidJwk(f"Unsupported JWK param '{name}'")
            elif param.kind == "name":
                pass
            else:
                assert (
                    False
                ), f"Unsupported param '{name}' type '{param.kind}'"  # pragma: no cover

        # if at least one of the supplied parameter was private, then all required private parameters must be provided
        if jwk_is_private:
            for name, param in self.PARAMS.items():
                value = self.get(name)
                if param.is_private and param.is_required and value is None:
                    raise InvalidJwk(
                        f"Missing required private param {param.description} ({name})"
                    )

        # if key is used for signing, it must be private
        for op in self.get("key_ops", []):
            if op in ("sign", "decrypt", "unwrapKey") and not self.is_private:
                raise InvalidJwk(f"Key Operation is '{op}' but the key is public")

    def supported_signing_algorithms(self) -> List[str]:
        """Return the list of Signature algorithms that can be used with this key.

        Returns:
          a list of supported algs
        """
        return list(self.SIGNATURE_ALGORITHMS)

    def supported_key_management_algorithms(self) -> List[str]:
        """Return the list of Key Management algorithms that can be used with this key.

        Returns:
            a list of supported algs
        """
        return list(self.KEY_MANAGEMENT_ALGORITHMS)

    def supported_encryption_algorithms(self) -> List[str]:
        """Return the list of Encryption algorithms that can be used with this key.

        Returns:
            a list of supported algs
        """
        return list(self.ENCRYPTION_ALGORITHMS)

    def public_jwk(self) -> Jwk:
        """Return the public Jwk associated with this key.

        Returns:
          a Jwk with the public key
        """
        if not self.is_private:
            return self

        params = {
            name: self.get(name)
            for name, param in self.PARAMS.items()
            if not param.is_private
        }

        key_ops = self.get("key_ops")
        if key_ops:
            if "sign" in key_ops:
                key_ops.remove("sign")
                key_ops.append("verify")
            if "decrypt" in key_ops:
                key_ops.remove("decrypt")
                key_ops.append("encrypt")
            if "unwrapKey" in key_ops:
                key_ops.remove("unwrapKey")
                key_ops.append("wrapKey")

        return Jwk(
            dict(
                kty=self.kty,
                kid=self.get("kid"),
                alg=self.get("alg"),
                use=self.get("use"),
                key_ops=key_ops,
                **params,
            )
        )

    def as_jwks(self) -> JwkSet:
        """Return a JwkSet with this key as single element.

        Returns:
            a JwsSet with this single key
        """
        from .jwks import JwkSet

        return JwkSet(keys=(self,))

    def sign(self, data: bytes, alg: Optional[str] = None) -> BinaPy:
        """Sign a data using this Jwk, and return the generated signature.

        Args:
          data: the data to sign
          alg: the alg to use (if this key doesn't have an `alg` parameter)

        Returns:
          the generated signature
        """
        sigalg = select_alg(self.alg, alg, self.SIGNATURE_ALGORITHMS)
        wrapper: BaseSignatureAlg
        if issubclass(sigalg, BaseAsymmetricAlg):
            wrapper = sigalg(self.cryptography_key)

        elif issubclass(sigalg, BaseSymmetricAlg):
            wrapper = sigalg(self.key)

        signature = wrapper.sign(data)
        return BinaPy(signature)

    def verify(
        self,
        data: bytes,
        signature: bytes,
        *,
        alg: Optional[str] = None,
        algs: Optional[Iterable[str]] = None,
    ) -> bool:
        """Verify a signature using this Jwk, and return `True` if valid.

        Args:
          data: the data to verify
          signature: the signature to verify
          alg: the allowed signature alg, if there is only one
          algs: the allowed signature algs, if there are several

        Returns:
          `True` if the signature matches, `False` otherwise
        """
        wrapper: BaseSignatureAlg
        for sigalg in select_algs(self.alg, alg, algs, self.SIGNATURE_ALGORITHMS):
            if issubclass(sigalg, BaseAsymmetricAlg):
                key = self.public_jwk().cryptography_key
                wrapper = sigalg(key)
            elif issubclass(sigalg, BaseSymmetricAlg):
                key = self.key
                wrapper = sigalg(key)
            if wrapper.verify(data, signature):
                return True

        return False

    def encrypt(
        self,
        plaintext: bytes,
        *,
        aad: Optional[bytes] = None,
        alg: Optional[str] = None,
        iv: Optional[bytes] = None,
    ) -> Tuple[BinaPy, BinaPy, BinaPy]:
        """Encrypt a plaintext, with an optional Additional Authenticated Data (AAD) using this JWK, and return the Encrypted Data, the Initialization Vector and the Authentication Tag.

        Args:
          plaintext: the data to encrypt.
          aad: the Additional Authenticated Data (AAD) to include in the authentication tag
          alg: the alg to use to encrypt the data
          iv: the Initialization Vector that was used to encrypt the data. If `iv` is passed as parameter, this
        will return that same value. Otherwise, an IV is generated.

        Returns:
          a tuple (ciphertext, iv, authentication_tag), as raw data
        """
        raise NotImplementedError  # pragma: no cover

    def decrypt(
        self,
        ciphertext: bytes,
        *,
        iv: bytes,
        tag: bytes,
        aad: Optional[bytes] = None,
        alg: Optional[str] = None,
    ) -> BinaPy:
        """Decrypt an encrypted data using this Jwk, and return the encrypted result.

        This is implemented by subclasses.

        Args:
          ciphertext: the data to decrypt
          iv: the Initialization Vector (IV) that was used for encryption
          tag: the Authentication Tag that will be verified while decrypting data
          aad: the Additional Authentication Data (AAD) to verify the Tag against
          alg: the alg to use for decryption

        Returns:
          the clear-text data
        """
        raise NotImplementedError  # pragma: no cover

    def sender_key(
        self,
        *,
        enc: str,
        alg: Optional[str],
        cek: Optional[bytes] = None,
        epk: Optional[Jwk] = None,
        **headers: Any,
    ) -> Tuple[Jwk, BinaPy, Mapping[str, Any]]:
        """For DH-based algs. As a token issuer, derive a EPK and CEK from the recipient public key.

        Returns a tuple with 3 items:

        - the clear text CEK, as a SymmetricJwk instance. Use this key to encrypt your message, but do not communicate this key!
        - the encrypted CEK, as bytes. You must send this to your recipient. This may be empty for algs which derive a CEK instead of generating one.
        - extra headers depending on the Key Management algorithm, as a dict of name to values: you must send this to your recipient as well.

        For algorithms that rely on a randomly generated CEK, you can provide that value instead of letting `jwskate` generate a
        safe, unique random value for you. Likewise, for algorithms that rely on an ephemeral key, you can provide an
        EPK that you generated yourself, instead of letting `jwskate` generate an appropriate value for you.
        Only use this if you know what you are doing!

        Args:
          enc: the encryption algorithm to use with the CEK
          alg: the Key Management algorithm to use to produce the CEK
          cek: CEK to use (leave `None` to have an adequate random value generated automatically)
          epk: EPK to use (leave `None` to have an adequate ephemeral key generated automatically)
          **headers: additional headers to include for the CEK derivation

        Returns:
          Tuple[Jwk,BinaPy,Mapping[str,Any]]: a tuple (cek, wrapped_cek, additional_headers_map)

        Raises:
            UnsupportedAlg: if the requested alg identifier is not supported
        """
        from jwskate import SymmetricJwk

        keyalg = select_alg(self.alg, alg, self.KEY_MANAGEMENT_ALGORITHMS)
        encalg = select_alg(None, enc, SymmetricJwk.ENCRYPTION_ALGORITHMS)

        cek_headers: Dict[str, Any] = {}

        if issubclass(keyalg, BaseRsaKeyWrap):
            rsa: BaseRsaKeyWrap = keyalg(self.public_jwk().cryptography_key)
            if cek:
                encalg.check_key(cek)
            else:
                cek = encalg.generate_key()
            wrapped_cek = rsa.wrap_key(cek)

        elif issubclass(keyalg, EcdhEs):
            ecdh: EcdhEs = keyalg(self.public_jwk().cryptography_key)
            epk = epk or Jwk.from_cryptography_key(ecdh.generate_ephemeral_key())
            cek_headers = {"epk": epk.public_jwk()}
            if isinstance(ecdh, BaseEcdhEs_AesKw):
                if cek:
                    encalg.check_key(cek)
                else:
                    cek = encalg.generate_key()
                wrapped_cek = ecdh.wrap_key_with_epk(
                    cek, epk.cryptography_key, alg=alg, **headers
                )
            else:
                cek = ecdh.sender_key(
                    epk.cryptography_key,
                    alg=encalg.name,
                    key_size=encalg.key_size,
                    **headers,
                )
                wrapped_cek = BinaPy(b"")
        elif issubclass(keyalg, BaseAesKeyWrap):
            aes: BaseAesKeyWrap = keyalg(self.cryptography_key)
            if cek:
                encalg.check_key(cek)
            else:
                cek = encalg.generate_key()
            wrapped_cek = aes.wrap_key(cek)

        elif issubclass(keyalg, BaseAesGcmKeyWrap):
            aesgcm: BaseAesGcmKeyWrap = keyalg(self.cryptography_key)
            if cek:
                encalg.check_key(cek)
            else:
                cek = encalg.generate_key()
            iv = aesgcm.generate_iv()
            wrapped_cek, tag = aesgcm.wrap_key(cek, iv=iv)
            cek_headers = {
                "iv": iv.to("b64u").ascii(),
                "tag": tag.to("b64u").ascii(),
            }

        elif issubclass(keyalg, DirectKeyUse):
            dir: DirectKeyUse = keyalg(self.key)
            cek = dir.direct_key(encalg)
            wrapped_cek = BinaPy(b"")
        else:
            raise UnsupportedAlg(f"Unsupported Key Management Alg {keyalg}")

        return SymmetricJwk.from_bytes(cek), wrapped_cek, cek_headers

    def recipient_key(
        self, wrapped_cek: bytes, *, alg: str, enc: str, **headers: Any
    ) -> Jwk:
        """For DH-based algs. As a token recipient, derive the same CEK that was used for encryption, based on the recipient private key and the sender ephemeral public key.

        Args:
          wrapped_cek: the wrapped CEK
          alg: the Key Management algorithm to use to unwrap the CEK
          enc: the encryption algorithm to use with the CEK
          **headers:

        Returns:
          the clear-text CEK, as a SymmetricJwk instance

        Raises:
            UnsupportedAlg: if the requested alg identifier is not supported
        """
        from jwskate import SymmetricJwk

        keyalg = select_alg(self.alg, alg, self.KEY_MANAGEMENT_ALGORITHMS)
        encalg = select_alg(None, enc, SymmetricJwk.ENCRYPTION_ALGORITHMS)

        if issubclass(keyalg, BaseRsaKeyWrap):
            rsa = keyalg(self.cryptography_key)
            cek = rsa.unwrap_key(wrapped_cek)

        elif issubclass(keyalg, EcdhEs):
            ecdh = keyalg(self.cryptography_key)
            epk = headers.get("epk")
            if epk is None:
                raise ValueError("No EPK in the headers!")
            epk_jwk = Jwk(epk)
            if epk_jwk.is_private:
                raise ValueError("The EPK present in the header is private.")
            epk = epk_jwk.cryptography_key
            encalg = select_alg(None, enc, SymmetricJwk.ENCRYPTION_ALGORITHMS)
            if isinstance(ecdh, BaseEcdhEs_AesKw):
                cek = ecdh.unwrap_key_with_epk(wrapped_cek, epk, alg=alg)
            else:
                cek = ecdh.recipient_key(
                    epk, alg=encalg.name, key_size=encalg.key_size, **headers
                )

        elif issubclass(keyalg, BaseAesKeyWrap):
            aes = keyalg(self.cryptography_key)
            cek = aes.unwrap_key(wrapped_cek)

        elif issubclass(keyalg, BaseAesGcmKeyWrap):
            aesgcm = keyalg(self.cryptography_key)
            iv = headers.get("iv")
            if iv is None:
                raise ValueError("No 'iv' in headers!")
            iv = BinaPy(iv).decode_from("b64u")
            tag = headers.get("tag")
            if tag is None:
                raise ValueError("No 'tag' in headers!")
            tag = BinaPy(tag).decode_from("b64u")
            cek = aesgcm.unwrap_key(wrapped_cek, tag=tag, iv=iv)

        elif issubclass(keyalg, DirectKeyUse):
            dir_ = keyalg(self.key)
            cek = dir_.direct_key(encalg)
        else:
            raise UnsupportedAlg(f"Unsupported Key Management Alg {keyalg}")

        return SymmetricJwk.from_bytes(cek)

    @classmethod
    def from_cryptography_key(cls, cryptography_key: Any, **kwargs: Any) -> Jwk:
        """Initialize a Jwk from a key from the `cryptography` library.

        The input key can be any private or public key supported by cryptography.

        Args:
          cryptography_key: a `cryptography` key instance
          **kwargs: additional members to include in the Jwk (e.g. kid, use)

        Returns:
            the matching `Jwk` instance

        Raises:
            TypeError: if the key type is not supported
        """
        for klass in cryptography_key.__class__.mro():
            jwk_class = cls.cryptography_key_types.get(klass)
            if jwk_class:
                return jwk_class.from_cryptography_key(cryptography_key, **kwargs)

        raise TypeError(f"Unsupported Jwk class for this Key Type: {cryptography_key}")

    def _to_cryptography_key(self) -> Any:
        """Return a key from the `cryptography` library that matches this Jwk.

        This is implemented by subclasses.

        Returns:
            a `cryptography`key instance initialized from the current key
        """
        raise NotImplementedError

    @classmethod
    def from_pem_key(cls, data: bytes, password: Optional[bytes] = None) -> Jwk:
        """Load a Jwk from a PEM encoded private or public key.

        Args:
          data: the PEM encoded data to load
          password: the password to decrypt the PEM, if required

        Returns:
            a Jwk instance from the loaded key
        """
        try:
            cryptography_key = serialization.load_pem_private_key(data, password)
        except Exception:
            try:
                cryptography_key = serialization.load_pem_public_key(data)
                if password is not None:
                    raise ValueError(
                        "A public key was loaded from PEM, while a password was provided for decryption."
                        "Only private keys are encrypted in PEM."
                    )
            except Exception:
                raise ValueError(
                    "The provided data is not a private or a public PEM encoded key."
                )

        return cls.from_cryptography_key(cryptography_key)

    def to_pem_key(self, password: Optional[bytes] = None) -> str:
        """Serialize this key to PEM format.

        For private keys, you can provide a password for encryption.

        Args:
          password: password to use to encrypt the PEM

        Returns:
            the PEM encrypted key, as
        """
        raise NotImplementedError

    @classmethod
    def generate(cls, **kwargs: Any) -> Jwk:
        """Generates a Private Key. This method is implemented by subclasses for specific Key Types and returns an instance of that specific subclass.

        Args:
          **kwargs: specific parameters depending on the type of key, or additional members to include in the Jwk

        Returns:
            a Jwk instance with a generated key
        """
        raise NotImplementedError

    @classmethod
    def generate_for_kty(cls, kty: str, **kwargs: Any) -> Jwk:
        """Generate a key with a specific type and return the resulting Jwk.

        Args:
          kty: key type to generate
          **kwargs: specific parameters depending on the key type, or additional members to include in the Jwk

        Returns:
            the resulting Jwk

        Raises:
            UnsupportedKeyType: if the key type is not supported
        """
        jwk_class = cls.subclasses.get(kty)
        if jwk_class is None:
            raise UnsupportedKeyType("Unsupported Key Type:", kty)
        return jwk_class.generate(**kwargs)

    @classmethod
    def generate_for_alg(cls, alg: str, **kwargs: Any) -> Jwk:
        """Generate a key for usage with a specific alg and return the resuting Jwk.

        Args:
            alg: a signature or key management alg
            **kwargs: specific parameters depending on the key type, or additional members to include in the Jwk

        Returns:
            the resulting Jwk
        """



"""This module implement base classes used by Signature, Encryption and Key Management JWA algorithms."""

from contextlib import contextmanager
from typing import Generic, Iterator, Optional, Tuple, Type, TypeVar, Union

from binapy import BinaPy


class PrivateKeyRequired(AttributeError):
    """Raised when a cryptographic operation requires a private key, and a public key has been provided instead."""


class PublicKeyRequired(AttributeError):
    """Raised when a cryptographic operation requires a public key, and a private key has been provided instead."""


class BaseAlg:
    """Base class for all algorithms.

    An algorithm has a `name` and a `description`, whose reference is here: https://www.iana.org/assignments/jose/jose.xhtml#web-signature-encryption-algorithms
    """

    name: str
    """Technical name of the algorithm"""
    description: str
    """Description of the algorithm (human readable)"""
    read_only: bool = False
    """For algs that are considered insecure, allow only signature verification or decryption of existing data, but don't allow new signatures or new encryptions."""

    def __repr__(self) -> str:
        """Use the name of the alg as repr."""
        return self.name


class BaseSymmetricAlg(BaseAlg):
    """Base class for Symmetric algorithms (using a raw bytes key)."""

    def __init__(self, key: bytes):
        """Initialize a Symmetric alg with a given key.

        Args:
            key: the key to use for cryptographic operations
        """
        self.check_key(key)
        self.key = key

    @classmethod
    def check_key(cls, key: bytes) -> None:
        """Check that a given key is suitable for this alg class.

        This raises an exception if the key is not suitable.
        This method must be implemented by subclasses as required.

        Args:
          key: the key to check for this alg class

        Returns:
          Returns `None`. Raises an exception if the key is not suitable
        """
        pass

    @classmethod
    def supports_key(cls, key: bytes) -> bool:
        """Return `True` if the given key is suitable for this alg class, or `False` otherwise.

        This is a convenience wrapper around `check_key(key)`.

        Args:
          key: the key to check for this alg class

        Returns:
          `True` if the key is suitable for this alg class, `False` otherwise
        """
        try:
            cls.check_key(key)
            return True
        except Exception:
            return False


Kpriv = TypeVar("Kpriv")
Kpub = TypeVar("Kpub")


class BaseAsymmetricAlg(Generic[Kpriv, Kpub], BaseAlg):
    """Base class for asymmetric algorithms. Those can be initialised with a private or public key.

    The available cryptographic operations will depend on the alg and
    the provided key type.
    """

    private_key_class: Union[Type[Kpriv], Tuple[Type[Kpriv], ...]]
    public_key_class: Union[Type[Kpub], Tuple[Type[Kpub], ...]]

    def __init__(self, key: Union[Kpriv, Kpub]):
        """Initialise an Asymmetric alg with either a private or a public key from the `cryptography` lib.

        Args:
            key: the key to use.
        """
        self.check_key(key)
        self.key = key

    @classmethod
    def check_key(cls, key: Union[Kpriv, Kpub]) -> None:
        """Check that a given key is suitable for this alg class.

        This must be implemented by subclasses as required.

        Args:
          key: the key to use.

        Returns:
          Returns None. Raises an exception if the key is not suitable.

        Raises:
            Exception: if the key is not suitable for use with this alg class
        """

    @classmethod
    def supports_key(cls, key: Union[Kpriv, Kpub]) -> bool:
        """Return `True` if the given key is suitable for this alg class, or `False` otherwise.

        This is a convenience wrapper around `check_key(key)`.

        Args:
          key: the key to check for this alg class

        Returns:
          `True` if the key is suitable for this alg class, `False` otherwise
        """
        try:
            cls.check_key(key)
            return True
        except Exception:
            return False

    @contextmanager
    def private_key_required(self) -> Iterator[Kpriv]:
        """A context manager that checks if this alg is initialised with a private key.

        Yields:
            the private key

        Raises:
            PrivateKeyRequired: if the configured key is not private
        """
        if not isinstance(self.key, self.private_key_class):
            raise PrivateKeyRequired()
        yield self.key  # type: ignore

    @contextmanager
    def public_key_required(self) -> Iterator[Kpub]:
        """A context manager that checks if this alg is initialised with a public key.

        Yields:
            The public key

        Raises:
            PublicKeyRequired: if the configured key is private
        """
        if not isinstance(self.key, self.public_key_class):
            raise PublicKeyRequired()
        yield self.key  # type: ignore


class BaseSignatureAlg(BaseAlg):
    """Base class for signature algorithms."""

    def sign(self, data: bytes) -> BinaPy:
        """Sign arbitrary data, return the signature.

        Args:
          data: raw data to sign

        Returns:
          the raw signature
        """
        raise NotImplementedError

    def verify(self, data: bytes, signature: bytes) -> bool:
        """Verify a signature against some data.

        Args:
          data: the raw data to verify
          signature: the raw signature

        Returns:
          `True` if the signature matches, `False` otherwise.
        """
        raise NotImplementedError


class BaseAESEncryptionAlg(BaseSymmetricAlg):
    """Base class for AES encryption algorithms."""

    key_size: int
    tag_size: int
    iv_size: int

    @classmethod
    def check_key(cls, key: bytes) -> None:
        """Check that a key is suitable for this algorithm.

        Args:
          key: the key to check

        Raises:
            ValueError: if the key is not suitable
        """
        if len(key) * 8 != cls.key_size:
            raise ValueError(
                f"This key size of {len(key) * 8} bits doesn't match the expected keysize of {cls.key_size} bits"
            )

    @classmethod
    def generate_key(cls) -> BinaPy:
        """Generate a key of an appropriate size for this AES alg subclass.

        Returns:
            a random AES key
        """
        return BinaPy.random_bits(cls.key_size)

    @classmethod
    def generate_iv(cls) -> BinaPy:
        """Generate an Initialisation Vector of the appropriate size.

        Returns:
            a random IV
        """
        return BinaPy.random_bits(cls.iv_size)

    def encrypt(
        self, plaintext: bytes, *, iv: bytes, aad: Optional[bytes]
    ) -> Tuple[BinaPy, BinaPy]:
        """Encrypt arbitrary data (`plaintext`) with the given Initialisation Vector (`iv`) and optional Additional Authentication Data (`aad`), return the ciphered text and authentication tag.

        Args:
          plaintext: the data to encrypt
          iv: the Initialisation Vector to use
          aad: the Additional Authentication Data

        Returns:
          a tuple of ciphered data and authentication tag
        """
        raise NotImplementedError

    def decrypt(
        self, ciphertext: bytes, *, iv: bytes, auth_tag: bytes, aad: Optional[bytes]
    ) -> BinaPy:
        """Decrypt a ciphertext with a given Initialisation Vector (iv) and optional Additional Authentication Data (aad), returns the resulting clear text.

        Args:
          ciphertext: the data to decrypt
          iv: the Initialisation Vector to use. Must be the same one used during encryption
          auth_tag: the authentication tag
          aad: the Additional Authentication Data. Must be the same one used during encryption

        Returns:
          the deciphered data
        """
        raise NotImplementedError


class BaseKeyManagementAlg(BaseAlg):
    """Base class for Key Management algorithms."""



"""This module contains base classes for all tokens types handled by `jwskate`."""
import json
from typing import Any, Dict, Type, TypeVar, Union


class BaseCompactToken:
    """Base class for all tokens in Compact representation.

    This includes JWS, JWE, and JWT tokens.
    """

    def __init__(self, value: Union[bytes, str], max_size: int = 16 * 1024):
        """Initialize a JW{S,E,T} from its string representation.

        Args:
            value: the string or bytes representation of this Jwt
            max_size: if the JWT length is larger than this value, raise a `ValueError`.
                This is to avoid JSON deserialization vulnerabilities.
        """
        if len(value) > max_size:
            raise ValueError(
                f"This JWT size exceeds {max_size} bytes, which is abnormally big. "
                "This size limit is made to avoid potential JSON deserialization vulnerabilities or issues. "
                "You can increase this limit by passing a different `max_size` value as parameter."
            )

        if not isinstance(value, bytes):
            value = value.encode("ascii")

        self.value = value
        self.headers: Dict[str, Any]

    def __eq__(self, other: Any) -> bool:
        """Check that a Jwt is equal to another.

        Works with other instances of `Jwt`, or with `str` or `bytes`.

        Args:
            other: the other token to compare with

        Returns:
            `True` if the other token has the same representation, `False` otherwise
        """
        if isinstance(other, BaseCompactToken):
            return self.value == other.value
        if isinstance(other, str):
            return self.value.decode() == other
        if isinstance(other, bytes):
            return self.value == other
        return super().__eq__(other)

    def get_header(self, name: str) -> Any:
        """Get a header from this Jwt.

        Args:
            name: the header name

        Returns:
            the header value
        """
        return self.headers.get(name)

    def __repr__(self) -> str:
        """Returns the `str` representation of this token."""
        return self.value.decode()

    def __bytes__(self) -> bytes:
        """Return the `bytes` representation of this token."""
        return self.value


D = TypeVar("D", bound="BaseJsonDict")


class BaseJsonDict(Dict[str, Any]):
    """Base class Jwk and tokens in JSON representation."""

    @classmethod
    def from_json(cls: Type[D], j: str) -> D:
        """Initialize an object based on a string containing a JSON representation.

        Args:
          j: the JSON to parse, still serialized

        Returns:
            the resulting object
        """
        return cls(json.loads(j))

    def to_json(self, *args: Any, **kwargs: Any) -> str:
        """Serialize the current object into a JSON representation.

        Args:
          *args: additional args for json.dumps()
          **kwargs: additional kwargs for json.dumps()

        Returns:
            a JSON representation of the current object
        """
        return json.dumps(self, *args, **kwargs)



"""This module exposes all Key Management algorithms available in `jwskate`."""

from .aesgcmkw import A128GCMKW, A192GCMKW, A256GCMKW, BaseAesGcmKeyWrap
from .aeskw import A128KW, A192KW, A256KW, BaseAesKeyWrap
from .dir import DirectKeyUse
from .ecdh import BaseEcdhEs_AesKw, EcdhEs, EcdhEs_A128KW, EcdhEs_A192KW, EcdhEs_A256KW
from .pbes2 import BasePbes2, Pbes2_HS256_A128KW, Pbes2_HS384_A192KW, Pbes2_HS512_A256KW
from .rsa import (
    BaseRsaKeyWrap,
    RsaEsOaep,
    RsaEsOaepSha256,
    RsaEsOaepSha384,
    RsaEsOaepSha512,
    RsaEsPcks1v1_5,
)

__all__ = [
    "A128GCMKW",
    "A192GCMKW",
    "A256GCMKW",
    "BaseAesGcmKeyWrap",
    "A128KW",
    "A192KW",
    "A256KW",
    "BaseAesKeyWrap",
    "DirectKeyUse",
    "EcdhEs",
    "BaseEcdhEs_AesKw",
    "EcdhEs_A128KW",
    "EcdhEs_A192KW",
    "EcdhEs_A256KW",
    "BasePbes2",
    "Pbes2_HS256_A128KW",
    "Pbes2_HS384_A192KW",
    "Pbes2_HS512_A256KW",
    "BaseRsaKeyWrap",
    "RsaEsOaep",
    "RsaEsOaepSha256",
    "RsaEsOaepSha384",
    "RsaEsOaepSha512",
    "RsaEsPcks1v1_5",
]



"""This module implements JWK representing Elliptic Curve keys."""

from __future__ import annotations

from typing import Any, List, Mapping, Union

from backports.cached_property import cached_property
from binapy import BinaPy
from cryptography.hazmat.primitives import asymmetric

from jwskate.jwa import (
    ES256,
    ES256K,
    ES384,
    ES512,
    P_256,
    P_384,
    P_521,
    EcdhEs,
    EcdhEs_A128KW,
    EcdhEs_A192KW,
    EcdhEs_A256KW,
    EllipticCurve,
    secp256k1,
)

from .base import Jwk, JwkParameter


class UnsupportedEllipticCurve(KeyError):
    """Raised when an unsupported Elliptic curve is requested."""


class ECJwk(Jwk):
    """Represent an Elliptic Curve Jwk, with `kty=EC`."""

    KTY = "EC"

    CRYPTOGRAPHY_KEY_CLASSES = (
        asymmetric.ec.EllipticCurvePrivateKey,
        asymmetric.ec.EllipticCurvePublicKey,
    )

    PARAMS: Mapping[str, JwkParameter] = {
        "crv": JwkParameter("Curve", is_private=False, is_required=True, kind="name"),
        "x": JwkParameter(
            "X Coordinate", is_private=False, is_required=True, kind="b64u"
        ),
        "y": JwkParameter(
            "Y Coordinate", is_private=False, is_required=True, kind="b64u"
        ),
        "d": JwkParameter(
            "ECC Private Key", is_private=True, is_required=True, kind="b64u"
        ),
    }

    CURVES: Mapping[str, EllipticCurve] = {
        curve.name: curve for curve in [P_256, P_384, P_521, secp256k1]
    }

    SIGNATURE_ALGORITHMS = {
        sigalg.name: sigalg for sigalg in [ES256, ES384, ES512, ES256K]
    }

    KEY_MANAGEMENT_ALGORITHMS = {
        keyalg.name: keyalg
        for keyalg in [EcdhEs, EcdhEs_A128KW, EcdhEs_A192KW, EcdhEs_A256KW]
    }

    @property
    def is_private(self) -> bool:  # noqa: D102
        return "d" in self

    def _validate(self) -> None:
        if not isinstance(self.crv, str) or self.crv not in self.CURVES:
            raise UnsupportedEllipticCurve(self.crv)
        super()._validate()

    @classmethod
    def get_curve(cls, crv: str) -> EllipticCurve:
        """Get the EllipticCurve instance for a given curve identifier.

        Args:
          crv: the curve identifier

        Returns:
            the matching EllipticCurve instance

        Raises:
            UnsupportedEllipticCurve: if the curve identifier is not supported
        """
        curve = cls.CURVES.get(crv)
        if curve is None:
            raise UnsupportedEllipticCurve(crv)
        return curve

    @property
    def curve(self) -> EllipticCurve:
        """Get the EllipticCurve instance for this key.

        Returns:
            the EllipticCurve instance
        """
        return self.get_curve(self.crv)

    @classmethod
    def public(cls, crv: str, x: int, y: int, **params: str) -> "ECJwk":
        """Initialize a public ECJwk from its public parameters.

        Args:
          crv: the curve to use
          x: the x coordinate
          y: the y coordinate
          **params: additional member to include in the Jwk

        Returns:
          an ECJwk initialized with the supplied parameters
        """
        coord_size = cls.get_curve(crv).coordinate_size
        return cls(
            dict(
                key="EC",
                crv=crv,
                x=BinaPy.from_int(x, length=coord_size).to("b64u"),
                y=BinaPy.from_int(y, length=coord_size).to("b64u"),
                **params,
            )
        )

    @classmethod
    def private(cls, crv: str, x: int, y: int, d: int, **params: Any) -> "ECJwk":
        """Initialize a private ECJwk from its private parameters.

        Args:
          crv: the curve to use
          x: the x coordinate
          y: the y coordinate
          d: the elliptic curve private key
          **params: additional members to include in the JWK

        Returns:
          an ECJWk initialized with the supplied parameters
        """
        coord_size = cls.get_curve(crv).coordinate_size
        return cls(
            dict(
                kty="EC",
                crv=crv,
                x=BinaPy.from_int(x, coord_size).to("b64u").ascii(),
                y=BinaPy.from_int(y, coord_size).to("b64u").ascii(),
                d=BinaPy.from_int(d, coord_size).to("b64u").ascii(),
                **params,
            )
        )

    @property
    def coordinate_size(self) -> int:
        """The coordinate size to use with the key curve.

        Returns:
          32, 48, or 66 (bits)
        """
        return self.curve.coordinate_size

    @classmethod
    def from_cryptography_key(cls, cryptography_key: Any, **kwargs: Any) -> ECJwk:
        """Initialize an ECJwk from a `cryptography` key.

        Args:
          cryptography_key: `cryptography` key
          **kwargs: additional members to include in the Jwk

        Returns:
            an ECJwk initialized from the provided `cryptography` key
        """
        parameters = EllipticCurve.get_jwk_parameters(cryptography_key)
        return cls(parameters)

    def _to_cryptography_key(
        self,
    ) -> Union[
        asymmetric.ec.EllipticCurvePrivateKey,
        asymmetric.ec.EllipticCurvePublicKey,
    ]:
        """Initialize a `cryptography` key based on this Jwk.

        Returns:
            an EllipticCurvePublicKey or EllipticCurvePrivateKey
        """
        if self.is_private:
            return asymmetric.ec.EllipticCurvePrivateNumbers(
                private_value=self.ecc_private_key,
                public_numbers=asymmetric.ec.EllipticCurvePublicNumbers(
                    x=self.x_coordinate,
                    y=self.y_coordinate,
                    curve=self.curve.cryptography_curve,
                ),
            ).private_key()
        else:
            return asymmetric.ec.EllipticCurvePublicNumbers(
                x=self.x_coordinate,
                y=self.y_coordinate,
                curve=self.curve.cryptography_curve,
            ).public_key()

    @classmethod
    def generate(cls, crv: str = "P-256", **params: str) -> "ECJwk":
        """Generates a random ECJwk.

        Args:
          crv: the curve to use
          **params:

        Returns:
          a generated ECJwk

        Raises:
            UnsupportedEllipticCurve: if the provided curve identifier is not supported.
        """
        curve = cls.get_curve(crv)
        if curve is None:
            raise UnsupportedEllipticCurve(crv)
        x, y, d = curve.generate()
        return cls.private(
            crv=crv,
            x=x,
            y=y,
            d=d,
            **params,
        )

    @cached_property
    def x_coordinate(self) -> int:
        """Return the x coordinate from this ECJwk.

        Returns:
         the x coordinate (from parameter `x`)
        """
        return BinaPy(self.x).decode_from("b64u").to_int()

    @cached_property
    def y_coordinate(self) -> int:
        """Return the y coordinate from this ECJwk.

        Returns:
            the y coordinate (from parameter `y`)
        """
        return BinaPy(self.y).decode_from("b64u").to_int()

    @cached_property
    def ecc_private_key(self) -> int:
        """Return the ECC private key from this ECJwk.

        Returns:
             the ECC private key (from parameter `d`)
        """
        return BinaPy(self.d).decode_from("b64u").to_int()

    def supported_signing_algorithms(self) -> List[str]:
        """Return the list of supported signature algorithms for this key.

        Returns:
            a list of supported algorithms identifiers
        """
        return [
            name
            for name, alg in self.SIGNATURE_ALGORITHMS.items()
            if alg.curve == self.curve
        ]

    def supported_key_management_algorithms(self) -> List[str]:
        """Return the list of supported Key Management algorithms for this key.

        Returns:
             a list of supported algorithms identifiers
        """
        return list(self.KEY_MANAGEMENT_ALGORITHMS)

    def supported_encryption_algorithms(self) -> List[str]:
        """Return the list of support Encryption algorithms for this key.

        Returns:
             a list of supported algorithms identifiers
        """
        return list(self.ENCRYPTION_ALGORITHMS)
